/*****************************************************
******************************************************/

import java.applet.Applet;
import java.lang.*;
import java.util.*;
import java.awt.*;
import java.applet.*;
import java.awt.event.*;

public class Tetris extends Applet
       implements MouseListener, KeyListener, FocusListener
{
  static final int TIEMPO = 1000;

  static final int QUIETO = 0;
  static final int ABAJO = 1;
  static final int DCHA = 2;
  static final int IZDA = 3;
  static final int ROTAR = 4;

  static final int I_INICIAL = 2;
  static final int J_INICIAL = 4;

  static final int NADA = 0;
  static final int OCUPADO = 1;
  static final Color claro = new Color(0xC66848);
  static final Color normal = new Color(0xB65838);
  static final Color oscuro = new Color(0x963818);
  static final Color fondo = new Color(0xE68868);
 
  private int[][] delta = {
   // Figura ****
   {-1,  0, +1, +2}, { 0,  0,  0,  0},
   { 0,  0,  0,  0}, {-2, -1,  0, +1},
   {-1,  0, +1, +2}, { 0,  0,  0,  0},
   { 0,  0,  0,  0}, {-2, -1,  0, +1},

   // Figura ***
   //         *
   { 0,  0,  0, +1}, {-1,  0, +1,  0},
   {-1,  0, +1,  0}, { 0,  0,  0, +1},
   { 0,  0,  0, -1}, {-1,  0, +1,  0},
   { 0,  0, -1, +1}, {-1,  0,  0,  0},
   // Figura  **
   //        **
   {+1, +1,  0,  0}, {-1,  0,  0, +1},
   {-1,  0,  0, +1}, { 0,  0, +1, +1},
   {+1, +1,  0,  0}, {-1,  0,  0, +1},
   {-1,  0,  0, +1}, { 0,  0, +1, +1},
   // Figura  **
   //          **
   { 0,  0, +1, +1}, {-1,  0,  0, +1},
   {-1,  0,  0, +1}, {+1,  0, +1,  0},
   { 0,  0, +1, +1}, {-1,  0,  0, +1},
   {-1,  0,  0, +1}, {+1,  0, +1,  0},
   // Figura  **
   //         **
   { 0,  0, +1, +1}, { 0, +1, +1,  0},
   { 0,  0, +1, +1}, { 0, +1, +1,  0},
   { 0,  0, +1, +1}, { 0, +1, +1,  0},
   { 0,  0, +1, +1}, { 0, +1, +1,  0},
   // Figura  * 
   //         ***
   {-1,  0,  0,  0}, {-1, -1,  0, +1},
   {+1, +1,  0, -1}, {-1,  0,  0,  0},
   { 0,  0,  0, +1}, {-1,  0, +1, +1},
   {-1, -1,  0, +1}, {+1,  0,  0,  0},
   // Figura    *
   //         ***
   { 0,  0,  0, -1}, {-1,  0, +1, +1},
   {-1, -1,  0, +1}, {-1,  0,  0,  0},
   {+1,  0,  0,  0}, {-1, -1,  0, +1},
   {-1,  0, +1, +1}, { 0,  0,  0, +1},
  };

  private boolean finJuego = false;
  private int[][] tablero = new int[17][10];
  private int UNIDAD = 10;
  private int numFilas = 0;

  private int ficha = 0;
  private int orientacion = 0;
  private int posI = 0;
  private int posJ = 0;
  private Reloj reloj;
  public int movimiento = QUIETO;

  private void traza(String txt) {
    System.out.println("Tetris: "+txt);
  }

  private int deltaI(int tipoFicha, int orientacion, int indice) {
    return delta[tipoFicha*8+orientacion*2][indice];
  }

  private int deltaJ(int tipoFicha, int orientacion, int indice) {
    return delta[tipoFicha*8+orientacion*2+1][indice];
  }

  private synchronized void nuevaFicha() {
    ficha = (int) ((double) 7*Math.random());
    orientacion = 0;
    posI = I_INICIAL; 
    posJ = J_INICIAL; 
    setMovimiento(QUIETO);
    if (puedeIr(posI, posJ)) ponerEn(posI, posJ);
    else {
      traza("Se acabo");
      reloj.stop();
      finJuego = true;
    }
  }

  public synchronized void rotar() {
    traza("Rotar");

    sacarDe(posI, posJ);
    int aux = orientacion;
    if (orientacion < 3) orientacion++;
    else orientacion = 0;
    if (puedeIr(posI, posJ) == false) orientacion = aux;
    ponerEn(posI, posJ);
    setMovimiento(QUIETO);
  }

  public synchronized void izquierda() {
    traza("Izquierda");
    sacarDe(posI, posJ);
    if (puedeIr(posI, posJ-1)) posJ--;
    ponerEn(posI, posJ);
    setMovimiento(QUIETO);
  }

  public synchronized void derecha() {
    traza("Derecha");
    sacarDe(posI, posJ);
    if (puedeIr(posI, posJ+1)) posJ++;
    ponerEn(posI, posJ);
    setMovimiento(QUIETO);
  }

  public synchronized void abajo() {
    traza("abajo");
    sacarDe(posI, posJ);
    if (puedeIr(posI+1, posJ)) {
      posI++;
      ponerEn(posI, posJ);
    } else {
      ponerEn(posI, posJ);
      revisaLinea();
      nuevaFicha();
    }
    setMovimiento(QUIETO);
  }

  public synchronized void abajoDeprisa() {
    traza("abajoDeprisa");
    sacarDe(posI, posJ);
    while (puedeIr(posI+1, posJ)) {
      posI++;
    }
    ponerEn(posI, posJ);
    revisaLinea();
    nuevaFicha();
    setMovimiento(QUIETO);
  }

  private void revisaLinea() {
    for(int i=16; i>0; ) {
      int nOcup = 0;
      for(int j=0; j<10; j++) {
        if (tablero[i][j] == OCUPADO) nOcup++;
      }
      if (nOcup == 10) {
        numFilas++;
        for(int k=i; k >0; k--) {
          for(int j=0; j<10; j++) {
            tablero[k][j] = tablero[k-1][j];
          }
        }
        repaint();
      } else i--;
    }
  }

  private void tableroInicial() {

    finJuego = false;

    for(int i=0; i<17; i++) {
      for(int j=0; j<10; j++) {
        tablero[i][j] = NADA;
      }
    }

    numFilas = 0;
    setMovimiento(QUIETO);

    nuevaFicha();

    repaint();
  }

  private boolean puedeIr(int i, int j) {
    int ni;
    int nj;

    try {
      for (int k=0; k < 4; k++) {
        ni = i + deltaI(ficha, orientacion, k);
        nj = j + deltaJ(ficha, orientacion, k);
        if (tablero[ni][nj] == OCUPADO) return false;
      }
      return true;
    } catch (Exception e) { return false; }
  }

  private void sacarDe(int i, int j) {
    int ni;
    int nj;

    Graphics g = getGraphics();

    for (int k=0; k < 4; k++) {
      ni = i + deltaI(ficha, orientacion, k);
      nj = j + deltaJ(ficha, orientacion, k);
      tablero[ni][nj] = NADA;
      pintaNada(g, nj*UNIDAD, ni*UNIDAD);
    }
  }

  private void ponerEn(int i, int j) {
    int ni;
    int nj;

    Graphics g = getGraphics();

    for (int k=0; k < 4; k++) {
      ni = i + deltaI(ficha, orientacion, k);
      nj = j + deltaJ(ficha, orientacion, k);
      tablero[ni][nj] = OCUPADO;
      pintaOcupado(g, nj*UNIDAD, ni*UNIDAD);
    }
  }

  public void init()
  {
    tableroInicial();
    addMouseListener(this);
    addKeyListener(this);
    addFocusListener(this);
  }

  private void arrancaReloj() {
    reloj = new Reloj(this, TIEMPO);
    reloj.start();
    traza("Start Reloj!");
  }

  private void paraReloj() {
    reloj.stop();
    traza("Stop Reloj!");
  }

  public void start()
  {
  }

  public void stop()
  {
  }

  public void paint(Graphics g){
    Dimension dim = getSize();

    UNIDAD = dim.height / 18;

    Image db = createImage(dim.width, dim.height);
    Graphics gdb = db.getGraphics();

    for(int i=0; i<17; i++) {
      for(int j=0; j<10; j++) {
        try {
          if (tablero[i][j] == OCUPADO) {
            pintaOcupado(gdb, j*UNIDAD, i*UNIDAD);
          } else {
            pintaNada(gdb, j*UNIDAD, i*UNIDAD);
          }
        } catch (Exception e) {}
      }
    }

    gdb.drawString("Filas completadas = "+numFilas, UNIDAD, 18*UNIDAD);
    g.drawImage(db,0,0,dim.width,dim.height, this);
  }

  private void pintaOcupado(Graphics g, int x, int y) {
    int w = UNIDAD;

    g.setColor(normal);
    g.fillRect(x+2, y+2, w-4, w-4);
    g.setColor(claro);
    g.drawLine(x, y, x+w-1, y);
    g.drawLine(x, y+1, x+w-2, y+1);
    g.drawLine(x, y, x, y+w-1);
    g.drawLine(x+1, y, x+1, y+w-2);
    g.setColor(oscuro);
    g.drawLine(x+w-1, y+w-1, x+w-1, y);
    g.drawLine(x+w-2, y+w-2, x+w-2, y-1);
    g.drawLine(x+w-1, y+w-1, x, y+w-1);
    g.drawLine(x+w-2, y+w-2, x-1, y+w-2);
  }

  private void pintaNada(Graphics g, int x, int y) {
    g.setColor(fondo);
    g.fillRect(x, y, UNIDAD, UNIDAD);
  }

  public synchronized void setMovimiento(int mov) {
    movimiento = mov;
  }

  public synchronized int getMovimiento() {
    return movimiento;
  }

  public void mouseReleased(MouseEvent e) {
  }

  public void mousePressed(MouseEvent e){
  }

  public void mouseEntered(MouseEvent e){
  }

  public void mouseExited(MouseEvent e){
  }

  public void mouseClicked(MouseEvent e){
    if (finJuego) {
      paraReloj();
      tableroInicial();
      arrancaReloj();
      return;
    }

    if ((e.getModifiers() & MouseEvent.SHIFT_MASK) ==
         MouseEvent.SHIFT_MASK) {
      setMovimiento(ROTAR);
    } else if ((e.getModifiers() & MouseEvent.CTRL_MASK) == 
               MouseEvent.CTRL_MASK) {
      setMovimiento(ABAJO);
    } else if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) == 
               MouseEvent.BUTTON1_MASK) {
      setMovimiento(IZDA);
    } else if ((e.getModifiers() & MouseEvent.BUTTON3_MASK) ==
               MouseEvent.BUTTON3_MASK) {
      setMovimiento(DCHA);
    }
  }

  public void keyTyped(KeyEvent e) {
  }

  public void keyPressed(KeyEvent e) {
    if (finJuego) return;
    if (e.getKeyCode() == KeyEvent.VK_SPACE) {
      setMovimiento(ROTAR);
    } else if (e.getKeyCode() == KeyEvent.VK_LEFT) {
      setMovimiento(IZDA);
    } else if (e.getKeyCode() == KeyEvent.VK_RIGHT) {
      setMovimiento(DCHA);
    } else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
      setMovimiento(ABAJO);
    } else if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
      paraReloj();
      tableroInicial();
      arrancaReloj();
    }
  }

  public void keyReleased(KeyEvent e) {
    traza("KeyReleased:"+e);
  }

  public void focusGained(FocusEvent e) {
    arrancaReloj();
  }

  public void focusLost(FocusEvent e) {
    paraReloj();
  }

}

class Reloj extends Thread {
  static final int VELOCIDAD = 2;
  Tetris tetris;
  int tiempo = 1000;

  public Reloj(Tetris tetris, int tiempo) {
    this.tetris = tetris;
    this.tiempo = tiempo;
  }

  public void run() {
    long antes;
    long ahora;
    int tgc = 10;

    antes = System.currentTimeMillis();
    while(true) {
      switch (tetris.getMovimiento()) {
        case Tetris.ABAJO : tetris.abajoDeprisa(); break;
        case Tetris.IZDA : tetris.izquierda(); break;
        case Tetris.DCHA : tetris.derecha(); break;
        case Tetris.ROTAR : tetris.rotar(); break;
      }
      ahora = System.currentTimeMillis();
      if (ahora - antes > tiempo) {
        if (--tgc < 0) { Runtime.getRuntime().gc(); tgc = 10; }
        tetris.abajo();
        if (tiempo > 100) tiempo -= VELOCIDAD;
        antes = ahora;
      }
    }
  }
}

